// The actual game finds its place in here.
// All sort of collision detection, user input and
// rendering is done here.
//
// The blocks (and the rotation procedure) can be found
// in the "Blocks" unit

unit Game;

interface

uses
	Blocks, // The blocks we use
	Types, // Common FreePaTris Types
	Crt; // Console functions

// Interface functions for GamestateManager
function DoGame(mstimepassed: comp; var ingame : TGameState) : TGamestateId;
function EnterGame() : TGameState;

implementation
	
// Checks whether any blocked block cell hits any blocked gamefield cell
// Parameters
//  - _gamefield: hit testing will take place in here
//  - _block: The tested block
function GamefieldHitTest(const _gamefield : TGameField; const _block : TBlock) : boolean;
var
	_x, _y : byte; // Counting variables
begin
	// Assuming there is no hit, until we prove better
	GamefieldHitTest := false;
	
	// Iterate over block content
	for _x := 0 to Length(_block.Cells[_block.Orientation]) - 1 do 
	begin
		for _y := 0 to Length(_block.Cells[_block.Orientation,_x]) - 1 do 
		begin
			// Check whether the block cell
            // 1. Is still inside the gamefield (empty cells may be outside)
            // 2. Hits anything inside the gamefield
			if ((_x + _block.X) >= 0) and ((_y + _block.Y) >= 0) and
               ((_x + _block.X) <= (GamefieldWidth - 1)) and
               ((_y + _block.Y) <= (GamefieldHeight - 1)) and
                _block.Cells[_block.Orientation,_x, _y] and
                _gamefield[_x + _block.X, _y + _block.Y].Blocked then
			begin
				// Thats a hit!
				GamefieldHitTest := true;
			end;
		end;
	end;
end;

// Moves a row of the gamefield down
// Parameters
//  - gamefield: The gamefield that will have its rows shifted
//  - row: The row to shift from on
procedure MoveGamefieldRowDown(var gamefield : TGameField; row : byte);
var
	_x, _y : byte; // Counting variables
begin
	// Move all rows down except the highest
	// Counting needs to be done from bottom to top, otherwise
	// the highest row would erase all lower rows.
	for _y := row downto 1 do
	begin
		for _x := 0 to GamefieldWidth - 1 do
		begin
			// Copy the higher cell into the lower cell
			gamefield[_x,_y + 1].Blocked := gamefield[_x,_y].Blocked;
                       gamefield[_x,_y + 1].Colour := gamefield[_x,_y].Colour;
		end;
	end;
	
	// Fill the highest line with empty cells
	for _x := 0 to GameFieldWidth -1 do
	begin
		gamefield[_x,0].Blocked := false;
		gamefield[_x,0].Colour := GamefieldBackground;
	end;
end;

// Removes any full rows from the gamefield and returns
// the number of removed rows
//
// Parameters
//  gamefield: The gamefield that will have rows compressed
// Calls
//  - MoveGamefieldRowDown
function CompressGamefield(var gamefield : TGameField) : byte;
var
	_x, _y : byte; // Counting variables
	rowfull : boolean;
	rowsdeleted : byte;
begin
	// Currently no rows deleted
	rowsdeleted := 0;

	// Iterate over gamefield
	for _y	:= 0 to GamefieldHeight - 1 do 
	begin
		// We assume a row to be filled until we can prove the opposite
		rowfull := true;
		for _x := 0 to GamefieldWidth - 1 do 
		begin
			if (not gamefield[_x, _y].Blocked) then
			begin
				// This row is not filled
				rowfull := false;
				break;
			end;
		end;

		// Was the row filled?
		if (rowfull) then
		begin
			// Remove the row above this a row down
			MoveGamefieldRowDown(gamefield, _y - 1);

			rowsdeleted := rowsdeleted + 1;
		end;
	end;

	// Return # deleted rows
	CompressGamefield := rowsdeleted;
end;

// Turns the block into a part of the gamefield
// and creates a new block
// Returns true if the new block does immediatly collide
//
// Parameters
//  ingame: The whole gamestate containing the gamefield and the block,
//          both (block and gamefield) may be altered
//
// Calls
//  - GameFieldHitTest
//  - MigrateBlockToGamefield
//  - CompressGamefield
//  - CreateRandomBlockId
function MigrateBlockToGamefield(var ingame : TGameState) : boolean;
var
	_x, _y : byte; // Counting variables;
begin
	// Iterate over block content
	for _x := 0 to Length(ingame.Block.Cells[ingame.Block.Orientation]) - 1 do
	begin
		for _y := 0 to Length(ingame.Block.Cells[ingame.Block.Orientation, _x]) - 1 do
		begin
			// Make sure we do not read out of bounds at the gamefield
			// This can happen if a block leaves a whole row of cells empty
			if ((_x + ingame.Block.X) < (GamefieldWidth)) and ((_y + ingame.Block.Y) < (GamefieldHeight)) then
			begin
				// Only "write" to gamefield cells if the block cell is filled
				if (ingame.Block.Cells[ingame.Block.Orientation,_x, _y]) then
				begin
					ingame.Gamefield[_x + ingame.Block.X, _y + ingame.Block.Y].Blocked := ingame.Block.Cells[ingame.Block.Orientation,_x, _y];
					ingame.Gamefield[_x + ingame.Block.X, _y + ingame.Block.Y].Colour := ingame.Block.Colour;
				end
			end;
		end;
	end;
	
	// Compress the gamefield, in case any rows were filled
 	ingame.RowsFilled := ingame.RowsFilled + CompressGamefield(ingame.Gamefield);

	// Create the new block
	ingame.Block := CreateBlock(ingame.NextBlock);
	// And randomize the next
	ingame.NextBlock := CreateRandomBlockId();

	// And return false, if collides with the gamefield
	MigrateBlockToGamefield := GamefieldHitTest(ingame.Gamefield, ingame.block);
end;

// Draws the horizontal decoration line
// with the correct width
// Parameters
//  corner: The char to use for the very right and very left
//  middle: The char to use in between
procedure DrawHorizontalDecorationLine(corner : char; middle : char);
var
	i : byte; // Counting
begin
	// Left corner
	write(corner);

	// Center bits
	for i := 1 to GamefieldWidth do
	begin
 		write(middle);
	end;

	// Right corner
	write(corner);

	// And newline
	writeln();
end;

// Draws the window decoration
//  - DrawHorizontalDecorationLine
procedure DrawDecoration();
var
	i : byte; // Counting
begin
	// Top
    DrawHorizontalDecorationLine('+', '-');
	
	// Middle
	for i := 1 to GameFieldHeight do
	begin
		DrawHorizontalDecorationLine('|', ' ');
	end;
	
	// Bottom
	DrawHorizontalDecorationLine('+', '-');
end;

// Draws the gamefield cells
//
// Parameters
//  gamefield: The gamefield that shal be drawn
procedure DrawGamefield(const gamefield : TGameField);
var
	_x, _y : byte; // Counting variables
begin
	// Iterate over gamefield and draw blocked cells
	for _x := 0 to GameFieldWidth - 1 do
	begin
		for _y := 0 to GameFieldHeight - 1 do
		begin
			gotoxy(_x + 2, _y + 2);
			if gamefield[_x, _y].Blocked then
			begin
				// Set correct background for cell
				textbackground(gamefield[_x, _y].Colour);
				write(DrawCharBlockedField);
			end else 
			begin
				textbackground(GamefieldBackground);
				write(DrawCharEmptyField);
			end;
		end;
	end;
end;

// Draws the falling block
//
// Parameters
//  block: The block that shal be drawn
procedure DrawBlock(_block : TBlock);
var
	_x, _y : byte; // Counting variables
begin
	// Set correct background for block
	textbackground(_block.Colour);
	
	// Iterate over block content
	for _x := 0 to Length(_block.Cells[_block.Orientation]) - 1 do begin
		for _y := 0 to Length(_block.Cells[_block.Orientation,_x]) - 1 do begin
			
			// Check wether the block content at the position is set
			if _block.Cells[_block.Orientation,_x, _y] then
			begin
				// Draw the block content
				// _block.X and Y are the offset of the block
				// _x and _Y are the offset of the blocks content
				gotoxy(_block.X + _x + 2, _block.Y + _y + 2);
				write(DrawCharFallingField);
			end;
		end;
	end;
	
	// Set the old background again
	textbackground(GamefieldBackground);
end;

// General drawing procedure
// Clears the screen and then draws decoration, gamefield, block and
// the number of cleared lines.
// If the DebugOutput constant is set to true, it also prints debug information.
//
// Parameters
//  mstimepassed: The frametime in ms
//  ingame: The whole gamestate containing the information we want to draw
//
// Calls
//  - DrawDecoration
//  - DrawGamefield
//  - DrawBlock
procedure Draw(mstimepassed: comp; const ingame : TGameState);
begin
	// Clear the screen before drawing
	clrscr();
	
	// Drawing the decoration
	DrawDecoration();
	
	// Drawing the gamefield
	DrawGamefield(ingame.gamefield);
	
	// Drawing the block
	DrawBlock(ingame.Block);

	// Write score and nextblock
	gotoxy(1,GamefieldHeight + 3);
	writeln('Rows:', ingame.RowsFilled, ' Lvl: ', ingame.RowsFilled div DefaultRowLevelStep);
	writeln('Next Block:', char(ingame.NextBlock));

	// Write debug output
	if (DebugOutput) then
	begin
		writeln('Upper:', ingame.Block.Boundary.Upper);
		writeln('Lower:', ingame.Block.Boundary.Lower);
		writeln('Left:', ingame.Block.Boundary.Left);
		writeln('Right:', ingame.Block.Boundary.Right);
		writeln('Orien:', ingame.Block.Orientation);
	end
end;

// Processes the user input and moves the falling block accordingly
//
// Parameters
//  ingame: The gamestate that shal be altered through the user input
//
// Calls
//  - GameFieldHitTest
//  - RotateBlock
function DoInput(var ingame : TGameState) : TGamestateId;
var
	shift : shortint;
	rotate : boolean;
	key : char;
begin
	shift := 0;
	rotate := false;
	DoInput := StateGame;

	// Check whether a key was pressed and test if it was a controlkey
	if (keypressed()) then
	begin
		key := readkey();
		if (key = #0) then
		begin
			// Check which controlkey was pressed
			key := readkey();
			case key of
				// Left
				#75:
					// Make sure the block does not move outside
					// at the left
					if (ingame.Block.X + ingame.Block.Boundary.Left) > 0 then
					begin
						// If it doesn't, move it left
						shift := - 1;
					end;
				// Right
				#77:
					// Make sure the block does not move outside
					// at the left
					if (ingame.Block.X + ingame.Block.Boundary.Right) < (GamefieldWidth - 1) then
					begin
						// If it doesn't, move it right
						shift := + 1;
					end;
				// Up
				#72:
					rotate := true;
                // F1
				#59:
					DoInput := StateMenu;
				#60:
					DoInput := StateQuit;
			end;
		end;
	end;
	
	// If there is anything to shift
	if (shift <> 0) then
	begin
		// Shift the block
		ingame.Block.X := ingame.Block.X + shift;
		// Check wether the new position is valid
		if (GamefieldHitTest(ingame.Gamefield, ingame.Block)) then
		begin
			// If not, move it back
			ingame.Block.X := ingame.Block.X - shift;
		end;
	end;
	
	// Is there anything to rotate?
	if (rotate) then
	begin
		// Rotate
		RotateBlock(RotateClockwise, ingame.Block);
		
		// Check for collision or boundaries
		if (GamefieldHitTest(ingame.Gamefield, ingame.Block)) or ((ingame.Block.X + ingame.Block.Boundary.Left) < 0) or
		   ((ingame.Block.X + ingame.Block.Boundary.Right) > GamefieldWidth - 1) then
		begin
			// If anything collided rotate back
			RotateBlock(not RotateClockwise, ingame.Block);
		end;
	end;
end;

// Returns the time a block needs to fall according to
// the level
//
// Parameters
//  ingame: The gamestate containing the scoring information from which the
//          falltime can be calculated
function CalculateFalltime(const ingame : TGameState) : comp;
var
	falltime : comp;
begin
	// Calculate falltime
    falltime := DefaultFallTime - ((ingame.RowsFilled div DefaultRowLevelStep) * DefaultFallTimeStep);

	// Make sure that blocks do not fall quicker then
	// 1 step per frame
	if (falltime < TargetFrametime) then
	begin
        falltime := TargetFrametime;
	end;

	CalculateFalltime := falltime;
end;

// Processes time, and routes user input and drawing
//
// Parameters
//  timepassed: the frametime in ms
//  ingame: The gamestate that shal be drawn and altered
//
// Calls
//  - DoInput
//  - Draw
//  - GamefieldHitTest
//  - MigrateBlockToGamefield
function DoGame(mstimepassed: comp; var ingame : TGameState) : TGamestateId;
	var initialcollision : boolean;
begin
	// We assume there is no initial collision until we prove better
	initialcollision := false;

	// Subtract the frametime from the redraw- and falltimer
	ingame.TimeTillDraw := ingame.TimeTillDraw - mstimepassed;
	ingame.TimeTillFall := ingame.TimeTillFall - mstimepassed;

	// Do general user input like block movement
	// and check wether the user wants to switch gamestate
	DoGame := DoInput(ingame);
	
	// If the redrawtimer is < 0, do the "normal" drawing procedures
	// and reset the timer
	if ingame.TimeTillDraw < 0 then
	begin
		ingame.TimeTillDraw := TargetFrametime;
		Draw(mstimepassed, ingame);
	end;
	
	// Do we need to move the block downwards?
	if ingame.TimeTillFall < 0 then
	begin
	
		// Check what happens when moving down
		// - We might touch the lower gamefield boundary
		if ((ingame.Block.Y + ingame.Block.Boundary.Lower) < GamefieldHeight - 1) then
		begin
			// Not touching, move down
			ingame.Block.Y := ingame.Block.Y + 1;
			
			// After moving, check for possible collision
			if GamefieldHitTest(ingame.Gamefield, ingame.Block) then
			begin
				// Collision! We move the block back up
				ingame.Block.Y := ingame.Block.Y - 1;
				
				// And migrate it into the gamefield
				initialcollision := MigrateBlockToGamefield(ingame);
			end;
		end else 
		begin
		 	initialcollision := MigrateBlockToGamefield(ingame);
		end;
		
		// Reset the falltimer
		ingame.TimeTillFall := CalculateFalltime(ingame);
	end;

	// Is there a initial collision?
	if (initialcollision) then
	begin
 		DoGame := StateMenu;
	end;
end;

// Starting a new game
//
// Returns the gamestate for the new game
//
// Calls
//  - CreateBlock
//  - CreateRandomBlockId
function EnterGame() : TGameState;
var
	_x, _y : byte; // Counting variables
	ingame : TGameState;
begin
	// Make sure the first frame calls the Draw() method
	ingame.TimeTillDraw := -1;
	ingame.RowsFilled := 0;
	ingame.TimeTillFall := CalculateFalltime(ingame);
	
	// Create the first block
	ingame.Block := CreateBlock(CreateRandomBlockId());

	// And tell about next block
	ingame.NextBlock := CreateRandomBlockId();
	
	// Initialise gamefield
	SetLength(ingame.GameField, GamefieldWidth, GamefieldHeight);
	
	// Iterate over gamefield and set default values
	for _x := 0 to Length(ingame.GameField) - 1 do
	begin
		for _y := 0 to Length(ingame.GameField[_x]) - 1 do
		begin
			ingame.GameField[_x, _y].Blocked := false;
			ingame.GameField[_x, _y].Colour := GamefieldBackground;
		end;
	end;

	// Return the created gamestate
	EnterGame := ingame;
end;

end.

